#!/usr/bin/perl -w
#
# treeConverter.pl v1.0
# Last modified January 26, 2005 16:11
# (c) Olaf R.P. Bininda-Emonds
#
# Input:
#	A nexus-foramtted treefile
#
# Output:
#   1) A standalone nexus-fromatted datafile
#	2) A phylip-formatted treefile
#
# Usage: treeConverter.pl  [-fn] [-fp] [-t<filename>] [-v] [-h]
#   options: -fn = output tree in standalone nexus format
#			 -fp = output tree in phylip format
#            -t<filename> = file containing trees in NEXUS format
#                           (default = treefile.tre)
#            -v = verbose output
#            -h = print this message and quit

use strict;

# Set user-input defaults and associated parameters
	my $block = 1;
		my $treeBlock = 0;
	my $treeFile = "treefile.tre";
		my $keepname = 1;
		my %root;
	my $nexus = 0;
	my $phylip = 0;
		my %phylipName;
	my $verbose = 0;

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] eq "-fn")
		{
		$nexus = 1;
		}
	elsif ($ARGV[$i] eq "-fp")
		{
		$phylip = 1;
		}
	elsif ($ARGV[$i] =~ /-t([\w+\.?\w+?]*)/)
		{
		$treeFile = $1;
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: treeConverter.pl [-fn] [-fp] [-t<filename>] [-v] [-h]\n";
		print "Options: -n = ignore tree names in treefile\n";
		print "         -fn = output tree in standalone nexus format\n";
		print "         -fp = output tree in phylip format\n";
		print "         -t<filename> = file containing trees in NEXUS format\n";
		print "                        (default = treefile.tre)\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: treeConverter.pl  [-n] [-fn] [-fp] [-t<filename>] [-v] [-h]\n";
		exit(1); 
		}
	}

die "ERROR: Must specify either a nexus- or phylip-formatted tree to be output\n" if (not $nexus and not $phylip);

(my $sourceName = $treeFile) =~ s/\.\w+$//;
	my $nexusFile = $sourceName . "_new.nex" if ($nexus);
	my $phylipFile = $sourceName . "_new.ph" if ($phylip);

setLineBreak($treeFile);
open (fTREES, "<$treeFile") or die "Cannot open treefile $treeFile\n";		# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	my (@AllTrees, %treeNames, $rawTreeDescrip, $treeDescrip, $treeFlag, $treeCount);
	my (@taxonList, %taxonPresent);
	my $NumChar = 0;
	my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
	
	while (<fTREES>)
		{
		chomp;
		$_ =~ s/^\s+//;		# Strip (any) leading white space
		if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
			{
			undef %TTable;
			$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
			$treeBlock++;
			}
		if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
			{
			$translateFlag = $translation = 1;
			$transText = "";
			print "Reading translation table ...\n" if ($verbose);
			}
		if ($translateFlag)
			{
			unless ($_ =~ m/translate/i)
				{
				$transText .= $_;
				if ($_ =~ m/;/)
					{
					$translateFlag = 0;
					$transText =~ s/;$//;	#Strip trailing semi-colon
					my @transEntries = split(/,/, $transText);
					foreach my $entry (@transEntries)
						{
						$entry =~ s/^\s+//;		#Strip (any) leading white space
						my ($taxonID, $taxonName) = split(/\s/,$entry);
						$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
						$taxonName =~ s/\s+//g;
						$TTable{$taxonID} = $taxonName;
						}
					}
				}
			}
		if ($_ =~ /tree/i and $_ =~ /=/)
			{
			unless ($_ =~ /^>/)		#Ignore (any) commented lines
				{
				$rawTreeDescrip = $treeDescrip = "";
				unless ($block == 0 and $treeBlock > 1)
					{
					$treeFlag = 1;
					}
				}
			}
		if ($treeFlag)
			{
			$rawTreeDescrip .= $_;
			if ($_ =~ m/;/)
				{
				$treeCount++;
				print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
				print "\tTranslating taxon names ...\n" if ($verbose and $translation);
				# Determine if trees are rooted or not (default for no information is not rooted)
					if ($rawTreeDescrip =~ /\[&R\]/)
						{
						$root{$treeCount} = 1;
						}
					else
						{
						$root{$treeCount} = 0;
						}
				# Remove any embedded comments
					if ($rawTreeDescrip =~ /\[/)
						{
						for (my $position = 0; $position < length($rawTreeDescrip); $position++)
							{
							if (substr($rawTreeDescrip, $position, 1) eq "[")
								{
								$position++ until (substr($rawTreeDescrip, $position, 1) eq "]");
								$position++;
								}
							else
								{
								$treeDescrip .= substr($rawTreeDescrip, $position, 1);
								}
							}
						}
					else
						{
						$treeDescrip = $rawTreeDescrip;
						}
				$treeFlag = 0;
				my $position = 0;
				until (substr($treeDescrip, $position, 1) eq "(")		#Avoid anything before tree description proper
					{
					$position++;
					}
				my $tree = substr($treeDescrip, $position);
					$tree =~ s/;\s*$//;
					$tree =~ s/\s+//g;
				my ($name) = (split /=/, $treeDescrip, 0);		#Get name of tree, if any
					$name =~ s/\s*\w*tree\s+//i;
					$name =~ s/\s*\*\s*//i;
					$name =~ s/\s+$//g;
					$treeNames{$treeCount} = $name;
				my $openCount = ($tree =~ tr/(//);
				my $closedCount = ($tree =~ tr/)//);
				if ($openCount == $closedCount)		#Tree is balanced; process
					{
					$position = length($tree) - 1;
					while ($position >= 0)
						{
						my $IDlength = 1;
						if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
							{
							$position--;
							}
						else	# Read in taxonID
							{
							until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
								{
								$position--;
								$IDlength++;
								}
							unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
								{
								my $readName = substr($tree, $position+1, $IDlength-1);
								if ($translation)	# Translate taxon names
									{
									if ($TTable{$readName})
										{
										substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
										unless ($taxonPresent{$TTable{$readName}})
											{
											push @taxonList, $TTable{$readName};
											$taxonPresent{$TTable{$readName}} = 1;
											}
										}
									else
										{
										print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
										substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
										unless ($taxonPresent{"xxx"})
											{
											push @taxonList, "xxx";
											$taxonPresent{"xxx"} = 1;
											}
										}
									}
								else	#Enter name as read from tree, unless already present
									{
									unless ($taxonPresent{$readName})
										{
										push @taxonList, $readName;
										$taxonPresent{$readName} = 1;
										}
									}
								}
							}
						}
					push @AllTrees, $tree;
				print "\tTree description:\n\t\t$tree\n" if ($verbose);
					}
				else	#Tree is unbalanced; reject
					{
					print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
					}
				}
			}
		}
close fTREES;

my @AllTaxa = sort {$a cmp $b} @taxonList;
	my $NumTaxa = scalar(@AllTaxa);
	
	if ($phylip)	# Trim taxon names to first 10 letters
		{
		print "\nThe following abbreviations were used for the phylip-formatted tree descriptions:\n" if ($verbose);
		my %shortNameCount;
		foreach my $name (@AllTaxa)
			{
			print "\t$name\t" if ($verbose);
			if (length($name) < 10)
				{
				$phylipName{$name} = $name;
				$shortNameCount{$name}++;
				print "$phylipName{$name}\n" if ($verbose);
				}
			else
				{
				my $trimmedName = substr($name, 0 , 10);
				$shortNameCount{$trimmedName}++;
				if ($shortNameCount{$trimmedName} > 1)	# Check for duplicates among shortened names and make unique by adding numbers
					{
					$phylipName{$name} = substr($name, 0, 10 - length($shortNameCount{$trimmedName}));
						$phylipName{$name} .= $shortNameCount{$trimmedName};
					}
				else
					{
					$phylipName{$name} = $trimmedName;
					}
				print "$phylipName{$name}\n" if ($verbose);
				}
			}
		}

# Print results
	if ($nexus)
		{
		open (NEX, ">$nexusFile") or die "Cannot write to nexus-formatted treefile, $nexusFile\n";
			print NEX "#nexus\n\n";
			print NEX "begin taxa;\n";
				print NEX "\tdimensions ntax = $NumTaxa;\n";
				print NEX "\ttaxlabels\n";
				print NEX "\t\t$_\n" foreach (@AllTaxa);
				print NEX "\t\t;\n";
				print NEX "end;\n\n";
			print NEX "begin trees;\n";
				for (1..$treeCount)
					{
					print NEX "\ttree $treeNames{$_} = ";
					if ($root{$_} == 1)
						{
						print NEX "[&R]";
						}
					else
						{
						print NEX "[&U]";
						}
					print NEX " $AllTrees[$_ - 1];\n";
					}
				print NEX "end;\n";
		close NEX;
		}
		
	if ($phylip)
		{
		open (PHYLIP, ">$phylipFile") or die "Cannot write to nexus-formatted treefile, $phylipFile\n";
			foreach my $tree (@AllTrees)
				{
				$tree =~ s/:\d+\.?\d*//g;
				foreach my $name (@AllTaxa)
					{
					next unless ($tree =~ /$name,/ or $tree =~ /$name\)/);
					$tree =~ s/$name,/$phylipName{$name},/;
					$tree =~ s/$name\)/$phylipName{$name}\)/;
					}
				print PHYLIP "$tree;\n";
				}
		close PHYLIP;
		}

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

# Version history
#
#	v1.0 (January 26, 2005)
#		- initial release
